@@ -1,3 +1,5 @@
+# frozen_string_literal: true
+
 require 'test_helper'
 
 class ZipFileTest < MiniTest::Test
@@ -8,19 +10,19 @@ class ZipFileTest < MiniTest::Test
   OK_DELETE_MOVED_FILE = 'test/data/generated/okToDeleteMoved.txt'
 
   def teardown
-    ::Zip.write_zip64_support = false
+    ::Zip.reset!
   end
 
   def test_create_from_scratch_to_buffer
     comment = 'a short comment'
 
-    buffer = ::Zip::File.add_buffer do |zf|
+    buffer = ::Zip::File.open_buffer(create: true) do |zf|
       zf.get_output_stream('myFile') { |os| os.write 'myFile contains just this' }
       zf.mkdir('dir1')
       zf.comment = comment
     end
 
-    ::File.open(EMPTY_FILENAME, 'wb') { |file| file.write buffer.string }
+    ::File.binwrite(EMPTY_FILENAME, buffer.string)
 
     zf_read = ::Zip::File.new(EMPTY_FILENAME)
     assert_equal(comment, zf_read.comment)
@@ -30,21 +32,7 @@ class ZipFileTest < MiniTest::Test
   def test_create_from_scratch
     comment = 'a short comment'
 
-    zf = ::Zip::File.new(EMPTY_FILENAME, ::Zip::File::CREATE)
-    zf.get_output_stream('myFile') { |os| os.write 'myFile contains just this' }
-    zf.mkdir('dir1')
-    zf.comment = comment
-    zf.close
-
-    zf_read = ::Zip::File.new(EMPTY_FILENAME)
-    assert_equal(comment, zf_read.comment)
-    assert_equal(2, zf_read.entries.length)
-  end
-
-  def test_create_from_scratch_with_old_create_parameter
-    comment = 'a short comment'
-
-    zf = ::Zip::File.new(EMPTY_FILENAME, 1)
+    zf = ::Zip::File.new(EMPTY_FILENAME, create: true)
     zf.get_output_stream('myFile') { |os| os.write 'myFile contains just this' }
     zf.mkdir('dir1')
     zf.comment = comment
@@ -57,7 +45,15 @@ class ZipFileTest < MiniTest::Test
 
   def test_get_input_stream_stored_with_gpflag_bit3
     ::Zip::File.open('test/data/gpbit3stored.zip') do |zf|
-      assert_equal("foo\n", zf.read('foo.txt'))
+      zis = zf.get_input_stream('file1.txt')
+
+      error = assert_raises(::Zip::StreamingError) do
+        zis.get_next_entry
+      end
+      assert_match(/file1\.txt/, error.message)
+      assert_equal('file1.txt', error.entry.name)
+
+      zf.get_input_stream('file2.txt')
     end
   end
 
@@ -71,27 +67,45 @@ class ZipFileTest < MiniTest::Test
       assert_equal(count + 1, zf.size)
       assert_equal('Putting stuff in new_entry.txt', zf.read('new_entry.txt'))
 
-      zf.get_output_stream(zf.get_entry('test/data/generated/empty.txt')) do |os|
-        os.write 'Putting stuff in data/generated/empty.txt'
-      end
+      # Use the non-block version of `get_output_stream` not tested elsewhere.
+      ostream =
+        zf.get_output_stream(zf.get_entry('test/data/generated/empty.txt'))
+      ostream.write 'Putting stuff in data/generated/empty.txt'
+      ostream.close
+
       assert_equal(count + 1, zf.size)
       assert_equal('Putting stuff in data/generated/empty.txt', zf.read('test/data/generated/empty.txt'))
 
-      custom_entry_args = [TEST_COMMENT, TEST_EXTRA, TEST_COMPRESSED_SIZE, TEST_CRC, ::Zip::Entry::STORED, TEST_SIZE, TEST_TIME]
-      zf.get_output_stream('entry_with_custom_args.txt', nil, *custom_entry_args) do |os|
+      custom_entry_args = {
+        comment: TEST_COMMENT, compressed_size: TEST_COMPRESSED_SIZE,
+        crc: TEST_CRC, compression_method: ::Zip::COMPRESSION_METHOD_STORE,
+        compression_level: ::Zlib::BEST_SPEED, size: TEST_SIZE, time: TEST_TIME
+      }
+      zf.get_output_stream(
+        'entry_with_custom_args.txt', **custom_entry_args
+      ) do |os|
         os.write 'Some data'
       end
+
       assert_equal(count + 2, zf.size)
       entry = zf.get_entry('entry_with_custom_args.txt')
-      assert_equal(custom_entry_args[0], entry.comment)
-      assert_equal(custom_entry_args[2], entry.compressed_size)
-      assert_equal(custom_entry_args[3], entry.crc)
-      assert_equal(custom_entry_args[4], entry.compression_method)
-      assert_equal(custom_entry_args[5], entry.size)
-      assert_equal(custom_entry_args[6], entry.time)
+      assert_equal(custom_entry_args[:comment], entry.comment)
+      assert_equal(custom_entry_args[:compressed_size], entry.compressed_size)
+      assert_equal(custom_entry_args[:crc], entry.crc)
+      assert_equal(
+        custom_entry_args[:compression_method], entry.compression_method
+      )
+      assert_equal(
+        custom_entry_args[:compression_level], entry.compression_level
+      )
+      assert_equal(custom_entry_args[:size], entry.size)
+
+      # Reverse times when testing because we need to use DOSTime#== for the
+      # comparison, not Time#==.
+      assert_equal(entry.time, custom_entry_args[:time])
 
       zf.get_output_stream('entry.bin') do |os|
-        os.write(::File.open('test/data/generated/5entry.zip', 'rb').read)
+        os.write(::File.binread('test/data/generated/5entry.zip'))
       end
     end
 
@@ -99,7 +113,34 @@ class ZipFileTest < MiniTest::Test
       assert_equal(count + 3, zf.size)
       assert_equal('Putting stuff in new_entry.txt', zf.read('new_entry.txt'))
       assert_equal('Putting stuff in data/generated/empty.txt', zf.read('test/data/generated/empty.txt'))
-      assert_equal(File.open('test/data/generated/5entry.zip', 'rb').read, zf.read('entry.bin'))
+      assert_equal(File.binread('test/data/generated/5entry.zip'), zf.read('entry.bin'))
+    end
+  end
+
+  def test_get_output_stream_with_entry
+    Dir.mktmpdir do |tmp|
+      test_zip = File.join(tmp, 'test.zip')
+      time = Time.new(1999, 12, 31)
+
+      ::Zip::File.open(test_zip, create: true) do |zip|
+        entry = ::Zip::Entry.new(zip.name, 'entry.txt', time: time)
+        zip.get_output_stream(entry) { |out| out.puts 'CONTENT!' }
+      end
+
+      ::Zip::File.open(test_zip) do |zip|
+        # Reverse times when testing because we need to use DOSTime#== for the
+        # comparison, not Time#==.
+        assert_equal(zip.get_entry('entry.txt').time, time)
+      end
+    end
+  end
+
+  def test_new_with_io_opened_non_binary_mode
+    File.open('test/data/test.xls') do |io|
+      refute(io.binmode?) # We open in non-binmode on purpose.
+      Zip::File.new(io, buffer: true) do |zip_io|
+        # left empty on purpose
+      end
     end
   end
 
@@ -128,29 +169,17 @@ class ZipFileTest < MiniTest::Test
   end
 
   def test_close_buffer_with_stringio
-    string_io = StringIO.new File.read('test/data/rubycode.zip')
+    string_io = StringIO.new File.read('test/data/rubycode.zip', mode: 'rb')
     zf = ::Zip::File.open_buffer string_io
     assert_nil zf.close
   end
 
-  def test_open_buffer_without_block_write_buffer_does_nothing
-    data = File.read('test/data/rubycode.zip', mode: 'rb')
-    string_io = StringIO.new(data.dup)
-
-    zf = ::Zip::File.open_buffer(string_io)
-    assert zf.entries.map(&:name).include?('zippedruby1.rb')
-
-    # Ensure the buffer isn't changed.
-    assert_same string_io, zf.write_buffer(string_io)
-    assert_equal(data, string_io.string)
-  end
-
   def test_open_buffer_no_op_does_not_change_file
     Dir.mktmpdir do |tmp|
       test_zip = File.join(tmp, 'test.zip')
       FileUtils.cp 'test/data/rubycode.zip', test_zip
 
-      # Note: this may change the file if it is opened with r+b instead of rb.
+      # NOTE: this may change the file if it is opened with r+b instead of rb.
       # The 'extra fields' in this particular zip file get reordered.
       File.open(test_zip, 'rb') do |file|
         Zip::File.open_buffer(file) do
@@ -183,7 +212,7 @@ class ZipFileTest < MiniTest::Test
 
   def test_open_buffer_with_io_and_block
     File.open('test/data/rubycode.zip') do |io|
-      io.set_encoding(Encoding::BINARY) # not strictly required but can be set
+      refute(io.binmode?) # We open in non-binmode on purpose.
       Zip::File.open_buffer(io) do |zip_io|
         # left empty on purpose
       end
@@ -191,13 +220,60 @@ class ZipFileTest < MiniTest::Test
   end
 
   def test_open_buffer_without_block
-    string_io = StringIO.new File.read('test/data/rubycode.zip')
+    string_io = StringIO.new File.read('test/data/rubycode.zip', mode: 'rb')
     zf = ::Zip::File.open_buffer string_io
     assert zf.entries.map(&:name).include?('zippedruby1.rb')
   end
 
+  def test_open_buffer_without_block_write_buffer_does_nothing
+    data = File.read('test/data/rubycode.zip', mode: 'rb')
+    string_io = StringIO.new(data.dup)
+
+    zf = ::Zip::File.open_buffer(string_io)
+    assert zf.entries.map(&:name).include?('zippedruby1.rb')
+
+    # Ensure the buffer isn't changed.
+    assert_same(string_io, zf.write_buffer(string_io))
+    assert_equal(data, string_io.string)
+  end
+
+  def test_open_file_with_max_length_comment
+    # Should not raise any errors.
+    Zip::File.open('test/data/max_length_file_comment.zip') do |zf|
+      assert_equal(1, zf.size)
+      assert_equal(65_535, zf.comment.length)
+    end
+  end
+
+  def test_open_zip64_file_with_max_length_comment
+    # Should not raise any errors.
+    Zip::File.open('test/data/zip64_max_length_file_comment.zip') do |zf|
+      assert_equal(2, zf.size)
+      assert_equal(65_535, zf.comment.length)
+    end
+  end
+
+  def test_count_entries
+    [
+      ['test/data/osx-archive.zip', 4],
+      ['test/data/zip64-sample.zip', 2],
+      ['test/data/max_length_file_comment.zip', 1],
+      ['test/data/100000-files.zip', 100_000]
+    ].each do |filename, num_entries|
+      assert_equal(num_entries, ::Zip::File.count_entries(filename))
+
+      ::File.open(filename, 'rb') do |f|
+        assert_equal(num_entries, ::Zip::File.count_entries(f))
+
+        f.seek(0)
+        s = StringIO.new(f.read)
+        assert_equal(num_entries, ::Zip::File.count_entries(s))
+      end
+    end
+  end
+
   def test_cleans_up_tempfiles_after_close
-    zf = ::Zip::File.new(EMPTY_FILENAME, ::Zip::File::CREATE)
+    zf = ::Zip::File.new(EMPTY_FILENAME, create: true)
     zf.get_output_stream('myFile') do |os|
       @tempfile_path = os.path
       os.write 'myFile contains just this'
@@ -210,27 +286,78 @@ class ZipFileTest < MiniTest::Test
     assert_equal(false, File.exist?(@tempfile_path))
   end
 
-  def test_add
+  def test_add_different_compression
     src_file = 'test/data/file2.txt'
     entry_name = 'newEntryName.rb'
-    assert(::File.exist?(src_file))
-    zf = ::Zip::File.new(EMPTY_FILENAME, ::Zip::File::CREATE)
-    zf.add(entry_name, src_file)
-    zf.close
+    files = [
+      ['test/data/fast_comp.zip', Zlib::BEST_SPEED],
+      ['test/data/default_comp.zip', Zlib::DEFAULT_COMPRESSION],
+      ['test/data/best_comp.zip', Zlib::BEST_COMPRESSION]
+    ]
+    sizes = []
 
-    zf_read = ::Zip::File.new(EMPTY_FILENAME)
-    assert_equal('', zf_read.comment)
-    assert_equal(1, zf_read.entries.length)
-    assert_equal(entry_name, zf_read.entries.first.name)
-    AssertEntry.assert_contents(src_file,
-                                zf_read.get_input_stream(entry_name, &:read))
+    files.each do |name, comp|
+      zf = ::Zip::File.new(name, create: true, compression_level: comp)
+
+      zf.add(entry_name, src_file)
+      zf.close
+
+      zf_read = ::Zip::File.new(name)
+      entry = zf_read.entries.first
+      assert_equal(File.size(src_file), entry.size)
+      refute(entry.zip64?) # No ZIP64 extra as we know the entry size here.
+      AssertEntry.assert_contents(
+        src_file, zf_read.get_input_stream(entry.name, &:read)
+      )
+      sizes << entry.compressed_size
+      zf_read.close
+
+      ::File.delete(name)
+    end
+
+    assert(sizes[0] > sizes[1])
+    assert(sizes[1] > sizes[2])
+  end
+
+  def test_add_different_compression_as_default
+    src_file = 'test/data/file2.txt'
+    entry_name = 'newEntryName.rb'
+    files = [
+      ['test/data/fast_comp.zip', Zlib::BEST_SPEED],
+      ['test/data/default_comp.zip', Zlib::DEFAULT_COMPRESSION],
+      ['test/data/best_comp.zip', Zlib::BEST_COMPRESSION]
+    ]
+    sizes = []
+
+    files.each do |name, comp|
+      ::Zip.default_compression = comp
+      zf = ::Zip::File.new(name, create: true)
+
+      zf.add(entry_name, src_file)
+      zf.close
+
+      zf_read = ::Zip::File.new(name)
+      entry = zf_read.entries.first
+      assert_equal(File.size(src_file), entry.size)
+      refute(entry.zip64?) # No ZIP64 extra as we know the entry size here.
+      AssertEntry.assert_contents(
+        src_file, zf_read.get_input_stream(entry.name, &:read)
+      )
+      sizes << entry.compressed_size
+      zf_read.close
+
+      ::File.delete(name)
+    end
+
+    assert(sizes[0] > sizes[1])
+    assert(sizes[1] > sizes[2])
   end
 
   def test_add_stored
     src_file = 'test/data/file2.txt'
     entry_name = 'newEntryName.rb'
     assert(::File.exist?(src_file))
-    zf = ::Zip::File.new(EMPTY_FILENAME, ::Zip::File::CREATE)
+    zf = ::Zip::File.new(EMPTY_FILENAME, create: true)
     zf.add_stored(entry_name, src_file)
     zf.close
 
@@ -242,29 +369,35 @@ class ZipFileTest < MiniTest::Test
     assert_equal(File.size(src_file), entry.size)
     assert_equal(entry.size, entry.compressed_size)
     assert_equal(::Zip::Entry::STORED, entry.compression_method)
+    refute(entry.zip64?) # No ZIP64 extra as we know the entry size here.
     AssertEntry.assert_contents(src_file,
                                 zf_read.get_input_stream(entry_name, &:read))
   end
 
   def test_recover_permissions_after_add_files_to_archive
+    # Windows NT does not support granular permissions
+    skip if Zip::RUNNING_ON_WINDOWS
+
     src_zip = TEST_ZIP.zip_name
-    ::File.chmod(0o664, src_zip)
-    src_file = 'test/data/file2.txt'
-    entry_name = 'newEntryName.rb'
-    assert_equal(::File.stat(src_zip).mode, 0o100664)
     assert(::File.exist?(src_zip))
-    zf = ::Zip::File.new(src_zip, ::Zip::File::CREATE)
-    zf.add(entry_name, src_file)
+
+    ::File.chmod(0o664, src_zip)
+    assert_equal(0o100664, ::File.stat(src_zip).mode)
+
+    zf = ::Zip::File.new(src_zip, create: true)
+    zf.add('newEntryName.rb', 'test/data/file2.txt')
     zf.close
-    assert_equal(::File.stat(src_zip).mode, 0o100664)
+
+    assert_equal(0o100664, ::File.stat(src_zip).mode)
   end
 
   def test_add_existing_entry_name
-    assert_raises(::Zip::EntryExistsError) do
+    error = assert_raises(::Zip::EntryExistsError) do
       ::Zip::File.open(TEST_ZIP.zip_name) do |zf|
         zf.add(zf.entries.first.name, 'test/data/file2.txt')
       end
     end
+    assert_match(/'add'/, error.message)
   end
 
   def test_add_existing_entry_name_replace
@@ -290,10 +423,50 @@ class ZipFileTest < MiniTest::Test
 
     ::Zip::File.open(TEST_ZIP.zip_name) do |zf|
       dir_entry = zf.entries.detect do |e|
-        e.name == TestFiles::EMPTY_TEST_DIR + '/'
+        e.name == "#{TestFiles::EMPTY_TEST_DIR}/"
       end
 
       assert(dir_entry.directory?)
+      refute(dir_entry.zip64?) # No ZIP64 extra as we know the entry size here.
+    end
+  end
+
+  def test_mkdir
+    buffer = ::Zip::File.open_buffer(create: true) do |zf|
+      # Add a directory with no slash.
+      zf.mkdir('dir')
+
+      # Add it again.
+      assert_raises(Errno::EEXIST) do
+        zf.mkdir('dir')
+      end
+
+      # Add it with a slash.
+      assert_raises(Errno::EEXIST) do
+        zf.mkdir('dir/')
+      end
+
+      # Add a directory with a slash.
+      zf.mkdir('folder/')
+
+      # Add it again.
+      assert_raises(Errno::EEXIST) do
+        zf.mkdir('folder/')
+      end
+
+      # Add it without a slash.
+      assert_raises(Errno::EEXIST) do
+        zf.mkdir('folder')
+      end
+    end
+
+    ::Zip::File.open_buffer(buffer) do |zf|
+      ['dir/', 'dir', 'folder/', 'folder'].each do |dir|
+        entry = zf.find_entry(dir)
+
+        assert(entry.directory?)
+        refute(entry.zip64?)
+      end
     end
   end
 
@@ -343,7 +516,7 @@ class ZipFileTest < MiniTest::Test
     ::File.unlink(zf_name) if ::File.exist?(zf_name)
     arr = []
     arr_renamed = []
-    ::Zip::File.open(zf_name, ::Zip::File::CREATE) do |zf|
+    ::Zip::File.open(zf_name, create: true) do |zf|
       zf.mkdir('test')
       arr << 'test/'
       arr_renamed << 'Ztest/'
@@ -356,7 +529,7 @@ class ZipFileTest < MiniTest::Test
     zf = ::Zip::File.open(zf_name)
     assert_equal(zf.entries.map(&:name), arr)
     zf.close
-    Zip::File.open(zf_name, 'wb') do |z|
+    Zip::File.open(zf_name) do |z|
       z.each do |f|
         z.rename(f, "Z#{f.name}")
       end
@@ -371,11 +544,12 @@ class ZipFileTest < MiniTest::Test
     old_entries = nil
     ::Zip::File.open(TEST_ZIP.zip_name) { |zf| old_entries = zf.entries }
 
-    assert_raises(::Zip::EntryExistsError) do
+    error = assert_raises(::Zip::EntryExistsError) do
       ::Zip::File.open(TEST_ZIP.zip_name) do |zf|
         zf.rename(zf.entries[0], zf.entries[1].name)
       end
     end
+    assert_match(/'rename'/, error.message)
 
     ::Zip::File.open(TEST_ZIP.zip_name) do |zf|
       assert_equal(old_entries.sort.map(&:name), zf.entries.sort.map(&:name))
@@ -419,7 +593,10 @@ class ZipFileTest < MiniTest::Test
   def test_rename_entry_to_existing_entry
     entry1, entry2, * = TEST_ZIP.entry_names
     zf = ::Zip::File.new(TEST_ZIP.zip_name)
-    assert_raises(::Zip::EntryExistsError) { zf.rename(entry1, entry2) }
+    error = assert_raises(::Zip::EntryExistsError) do
+      zf.rename(entry1, entry2)
+    end
+    assert_match(/'rename'/, error.message)
   ensure
     zf.close
   end
@@ -477,14 +654,27 @@ class ZipFileTest < MiniTest::Test
     assert_equal(res, true)
   end
 
+  def test_commit_preserves_options
+    zip_file = 'test/data/generated/preserve_options.zip'
+    ::Zip::File.open(zip_file, create: true, compression_level: 8) do |zf|
+      assert(zf.commit_required?)
+      zf.commit
+      assert_equal(8, zf.instance_variable_get(:@compression_level))
+      refute(zf.commit_required?)
+    end
+  end
+
   def test_double_commit(filename = 'test/data/generated/double_commit_test.zip')
     ::FileUtils.touch('test/data/generated/test_double_commit1.txt')
     ::FileUtils.touch('test/data/generated/test_double_commit2.txt')
-    zf = ::Zip::File.open(filename, ::Zip::File::CREATE)
+    zf = ::Zip::File.open(filename, create: true)
     zf.add('test1.txt', 'test/data/generated/test_double_commit1.txt')
     zf.commit
+    refute(zf.commit_required?)
     zf.add('test2.txt', 'test/data/generated/test_double_commit2.txt')
+    assert(zf.commit_required?)
     zf.commit
+    refute(zf.commit_required?)
     zf.close
     zf2 = ::Zip::File.open(filename)
     refute_nil(zf2.entries.detect { |e| e.name == 'test1.txt' })
@@ -494,7 +684,6 @@ class ZipFileTest < MiniTest::Test
   end
 
   def test_double_commit_zip64
-    ::Zip.write_zip64_support = true
     test_double_commit('test/data/generated/double_commit_test64.zip')
   end
 
@@ -503,9 +692,8 @@ class ZipFileTest < MiniTest::Test
     zf = ::Zip::File.new(TEST_ZIP.zip_name)
     old_name = zf.entries.first
     zf.rename(old_name, new_name)
-    io = ::StringIO.new
-    buffer = zf.write_buffer(io)
-    File.open(TEST_ZIP.zip_name, 'wb') { |f| f.write buffer.string }
+    buffer = zf.write_buffer(::StringIO.new)
+    File.binwrite(TEST_ZIP.zip_name, buffer.string)
     zf_read = ::Zip::File.new(TEST_ZIP.zip_name)
     refute_nil(zf_read.entries.detect { |e| e.name == new_name })
     assert_nil(zf_read.entries.detect { |e| e.name == old_name })
@@ -652,9 +840,9 @@ class ZipFileTest < MiniTest::Test
   end
 
   def test_streaming
-    fname = ::File.join(__dir__, '../README.md')
+    fname = ::File.join(__dir__, '..', 'README.md')
     zname = 'test/data/generated/README.zip'
-    Zip::File.open(zname, Zip::File::CREATE) do |zipfile|
+    Zip::File.open(zname, create: true) do |zipfile|
       zipfile.get_output_stream(File.basename(fname)) do |f|
         f.puts File.read(fname)
       end
@@ -664,14 +852,14 @@ class ZipFileTest < MiniTest::Test
     File.open(zname, 'rb') do |f|
       Zip::File.open_buffer(f) do |zipfile|
         zipfile.each do |entry|
-          next unless entry.name =~ /README.md/
+          next unless entry.name.include?('README.md')
 
           data = zipfile.read(entry)
         end
       end
     end
     assert data
-    assert data =~ /Simonov/
+    assert data.include?('Simonov')
   end
 
   def test_nonexistant_zip
